import scala.util.parsing.combinator._
import scala.util.parsing.combinator.syntactical.StdTokenParsers
import scala.util.parsing.combinator.lexical.StdLexical

abstract class Expr
case class Add(l: Expr, r: Expr) extends Expr
case class Sub(l: Expr, r: Expr) extends Expr
case class Mul(l: Expr, r: Expr) extends Expr
case class Div(l: Expr, r: Expr) extends Expr
case class Num(value: int) extends Expr
case class Assign(id: String, r: Expr) extends Expr
case class Ident(id: String) extends Expr
case class Program(body: List[Expr]) extends Expr

/**
 * Compiler for a simple language.
 */
object Compiler extends StdTokenParsers with ImplicitConversions with Application {
	type Tokens = StdLexical
	val lexical = new StdLexical
	lexical.delimiters ++= Set("(", ")", "+", "-", "*", "/", "=", ";")

    // BEGIN parse

    def program = ((expr ~ ";")+) ^^ ((e: List[Expr]) => Program(e))

    def expr: Parser[Expr] = assign | simpleExpr

    def assign = ident >> ((id: String) =>  "=" ~ expr ^^ (Assign(id, _)))

	def simpleExpr = term * (
	        "+" ^^ ((x: Expr, y: Expr) => Add(x, y))
	    |   "-" ^^ ((x: Expr, y: Expr) => Sub(x, y))
	)

	def term = factor * (
	        "*" ^^ ((x: Expr, y: Expr) => Mul(x, y))
	    |   "/" ^^ ((x: Expr, y: Expr) => Div(x, y))
	)

	def factor: Parser[Expr] = "(" ~ expr ~ ")" | ident ^^ ((s: String) => Ident(s)) | numericLit ^^ ((s: String) => Num(s.toInt))

    // END parse

    // BEGIN compile
    def generateCode(e: Expr): String = e match {
    	case Num(v) => "push " + v + "\n"
        case Ident(id) => pushString(id) + "deref\n"
    	case Mul(l, r) => generateCode(r) + generateCode(l) + "mul\n"
    	case Div(l, r) => generateCode(r) + generateCode(l) + "div\n"
    	case Add(l, r) => generateCode(r) + generateCode(l) + "add\n"
        case Sub(l, r) => generateCode(r) + generateCode(l) + "sub\n"
        case Assign(id, e) => generateCode(e) + pushString(id) + "bind\n"
        case Program(body) => body map(generateCode(_)) mkString ""
    }

    def pushString(s: String): String = "push '" + s + "'\n"
    // END compile

    program(new lexical.Scanner(Console.readLine())) match {
    	case Success(e, _) => Console.println(generateCode(e))
    	case Failure(msg, _) => Console.println("[FAILURE] " + msg)
    	case Error(msg, _) => Console.println("[ERROR] " + msg)
    }
}

